# -*- coding: utf-8 -*-

"""Test functions of solver.py"""

from __future__ import absolute_import

import numpy as np
import pytest

import cobra.util.solver as su
from cobra.exceptions import OptimizationError


stable_optlang = ["glpk", "cplex", "gurobi"]
optlang_solvers = ["optlang-" + s for s in stable_optlang if s in su.solvers]


def test_solver_list():
    assert len(su.solvers) >= 1
    assert "glpk" in su.solvers


def test_interface_str():
    assert su.interface_to_str("nonsense") == "nonsense"
    assert su.interface_to_str("optlang.glpk_interface") == "glpk"
    assert su.interface_to_str("optlang-cplex") == "cplex"


def test_solver_name():
    assert su.get_solver_name() == "glpk"


def test_choose_solver(model):
    so = su.choose_solver(model, "glpk")
    assert su.interface_to_str(so) == "glpk"

    if any(s in su.solvers for s in su.qp_solvers):
        qp_choice = su.choose_solver(model, qp=True)
        assert su.interface_to_str(qp_choice) in su.qp_solvers
    else:
        with pytest.raises(su.SolverNotFound):
            su.choose_solver(model, qp=True)


def test_linear_reaction_coefficients(model):
    coefficients = su.linear_reaction_coefficients(model)
    assert coefficients == {model.reactions.Biomass_Ecoli_core: 1}


@pytest.mark.parametrize("solver", optlang_solvers)
def test_fail_non_linear_reaction_coefficients(model, solver):
    model.solver = solver
    try:
        model.objective = model.problem.Objective(
            model.reactions.ATPM.flux_expression ** 2
        )
    except ValueError:
        pass
    else:
        coefficients = su.linear_reaction_coefficients(model)
        assert coefficients == {}
        with pytest.raises(ValueError):
            model.reactions.ACALD.objective_coefficient = 1


def test_add_remove(model):
    v = model.variables
    new_var = model.problem.Variable("test_var", lb=-10, ub=-10)
    new_constraint = model.problem.Constraint(
        v.PGK - new_var, name="test_constraint", lb=0)

    su.add_cons_vars_to_problem(model, [new_var, new_constraint])
    assert "test_var" in model.variables.keys()
    assert "test_constraint" in model.constraints.keys()

    su.remove_cons_vars_from_problem(model, [new_var, new_constraint])
    assert "test_var" not in model.variables.keys()
    assert "test_constraint" not in model.constraints.keys()


def test_add_remove_in_context(model):
    v = model.variables
    new_var = model.problem.Variable("test_var", lb=-10, ub=-10)

    with model:
        su.add_cons_vars_to_problem(model, [new_var])
        su.remove_cons_vars_from_problem(model, [v.PGM])
        assert "test_var" in model.variables.keys()
        assert "PGM" not in model.variables.keys()

    assert "test_var" not in model.variables.keys()
    assert "PGM" in model.variables.keys()


def test_absolute_expression(model):
    v = model.variables
    with model:
        parts = su.add_absolute_expression(
            model, 2 * v.PGM, name="test", ub=100)
        assert len(parts) == 3
        assert "test" in model.variables.keys()
        assert "abs_pos_test" in model.constraints.keys()
        assert "abs_neg_test" in model.constraints.keys()
    assert "test" not in model.variables.keys()
    assert "abs_pos_test" not in model.constraints.keys()
    assert "abs_neg_test" not in model.constraints.keys()


@pytest.mark.parametrize("solver", optlang_solvers)
def test_fix_objective_as_constraint(solver, model):
    model.solver = solver
    with model as m:
        su.fix_objective_as_constraint(model, 1.0)
        constraint_name = m.constraints[-1]
        assert abs(m.constraints[-1].expression -
                   m.objective.expression) < 1e-6
    assert constraint_name not in m.constraints
    su.fix_objective_as_constraint(model)
    constraint_name = model.constraints[-1]
    assert abs(model.constraints[-1].expression -
               model.objective.expression) < 1e-6
    assert constraint_name in model.constraints


@pytest.mark.parametrize("solver", optlang_solvers)
def test_fix_objective_as_constraint_minimize(model, solver):
    model.solver = solver
    model.reactions.Biomass_Ecoli_core.bounds = (0.1, 0.1)
    minimize_glucose = model.problem.Objective(
        model.reactions.EX_glc__D_e.flux_expression,
        direction='min')
    su.set_objective(model, minimize_glucose)
    su.fix_objective_as_constraint(model)
    fx_name = 'fixed_objective_{}'.format(model.objective.name)
    constr = model.constraints
    # Ensure that a solution exists on non-GLPK solvers.
    model.slim_optimize()
    assert (constr[fx_name].lb, constr[fx_name].ub) == (
        None, model.solver.objective.value)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_add_lp_feasibility(model, solver):
    model.solver = solver

    with model:
        with model:
            su.add_lp_feasibility(model)
            assert 's_plus_succoa_c' in model.variables
            assert np.isclose(model.slim_optimize(), 0.)

        model.reactions.EX_glc__D_e.lower_bound = 1
        assert np.isnan(model.slim_optimize())
        assert 's_plus_succoa_c' not in model.variables

        su.add_lp_feasibility(model)
        assert np.isclose(model.slim_optimize(), 1.3872307692307695)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_add_lexicographic_constraints(model, solver):
    model.solver = solver

    rxns = ['Biomass_Ecoli_core', 'EX_glc__D_e', 'EX_o2_e']

    with model:
        out = su.add_lexicographic_constraints(
            model, rxns, ['max', 'min', 'max'])
        print(model.reactions.Biomass_Ecoli_core.bounds)
        assert np.isclose(model.constraints[-3].lb, out[rxns[0]])
        assert np.isclose(model.constraints[-2].ub, out[rxns[1]])
        assert np.isclose(model.constraints[-1].lb, out[rxns[2]])

    with model:
        su.add_lexicographic_constraints(model, rxns, 'max')

    with model:
        su.add_lexicographic_constraints(model, rxns)


def test_time_limit(large_model):
    if su.interface_to_str(large_model.problem) != "glpk":
        pytest.skip("requires GLPK")

    # It is done like this since optlang accepts inputs in seconds
    # whereas GLPK accepts milliseconds
    large_model.solver.configuration._smcp.tm_lim = 1
    with pytest.warns(UserWarning):
        sol = large_model.optimize()
    assert sol.fluxes is not None

    with pytest.raises(OptimizationError):
        sol = large_model.optimize(raise_error=True)



# -*- coding: utf-8 -*-
#
# cobra documentation build configuration file, created by
# sphinx-quickstart on Wed Jun 13 19:17:34 2012.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
from os.path import dirname

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, dirname(dirname(__file__)))


# In order to build documentation that requires libraries to import
class Mock(object):
    def __init__(self, *args, **kwargs):
        return

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        else:
            return Mock()


# These modules should correspond to the importable Python packages.
MOCK_MODULES = [
    'depinfo',
    'numpy',
    'scipy', 'scipy.optimize', 'scipy.sparse', 'scipy.io', 'scipy.stats',
    'pp',
    'libsbml',
    'pandas',
    'tabulate',
    'optlang', 'optlang.interface', 'optlang.symbolics',
    'optlang.symbolics.core',
    'future', 'future.utils',
    'ruamel', 'ruamel.yaml'
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
    'autoapi.extension',
    'nbsphinx'
]
# Document Python Code
autoapi_type = 'python'
autoapi_dirs = ['..']
autoapi_ignore = ['.tox', '.pytest_cache', 'scripts', 'benchmarks']

# Napoleon settings
napoleon_numpy_docstring = True

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'cobra'
copyright = u'2016, The cobrapy core team'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
from cobra import __version__ as release
version = '.'.join(release.split('.')[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '.ipynb_checkpoints']

pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

mathjax_path = 'https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML'

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    'preamble': r'\usepackage{amsmath,amssymb}',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ('index', 'cobra.tex', u'cobra Documentation',
     u'The cobrapy core team', 'manual'),
]

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'cobra', u'cobra Documentation',
     [u'The cobrapy core team'], 1)
]

# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'cobra', u'cobra Documentation',
     u'The cobrapy core team', 'cobra',
     'A package for constraints-based modeling of biological networks',
     'Miscellaneous'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"http://docs.python.org/": None,
                       "http://docs.scipy.org/doc/numpy/": None,
                       "http://docs.scipy.org/doc/scipy/reference": None}
intersphinx_cache_limit = 10  # days to keep the cached inventories



# -*- coding: utf-8 -*-

"""Test functionalities of json.py"""

from __future__ import absolute_import

import json
from os.path import join

import pytest

import cobra.io as cio
from cobra.test.test_io.conftest import compare_models


@pytest.mark.xfail(reason="schema outdated")
def test_validate_json(data_directory):
    """Validate file according to JSON-schema."""
    jsonschema = pytest.importorskip("jsonschema")
    with open(join(data_directory, "mini.json"),
              "r", encoding="utf-8") as infile:
        loaded = json.load(infile)
    assert jsonschema.validate(loaded, cio.json.json_schema)


def test_load_json_model(data_directory, mini_model):
    """Test the reading of JSON model."""
    json_model = cio.load_json_model(join(data_directory, "mini.json"))
    assert compare_models(mini_model, json_model) is None


@pytest.mark.xfail(reason="schema outdated")
def test_save_json_model(tmpdir, mini_model):
    """Test the writing of JSON model."""
    jsonschema = pytest.importorskip("jsonschema")
    output_file = tmpdir.join("mini.json")
    cio.save_json_model(mini_model, output_file.strpath, pretty=True)
    # validate against JSONSchema
    with open(output_file, "r") as infile:
        loaded = json.load(infile)
    assert jsonschema.validate(loaded, cio.json.json_schema)



# -*- coding: utf-8 -*-

"""Test functionalities of MOMA."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.flux_analysis.moma import add_moma


def test_moma_sanity(model, qp_solvers):
    """Test optimization criterion and optimality for MOMA."""
    model.solver = qp_solvers
    sol = model.optimize()

    with model:
        model.reactions.PFK.knock_out()
        knock_sol = model.optimize()
        ssq = (knock_sol.fluxes - sol.fluxes).pow(2).sum()

    with model:
        add_moma(model, linear=False)
        model.reactions.PFK.knock_out()
        moma_sol = model.optimize()
        moma_ssq = (moma_sol.fluxes - sol.fluxes).pow(2).sum()

    # Use normal FBA as reference solution.
    with model:
        add_moma(model, solution=sol, linear=False)
        model.reactions.PFK.knock_out()
        moma_ref_sol = model.optimize()
        moma_ref_ssq = (moma_ref_sol.fluxes - sol.fluxes).pow(2).sum()

    assert np.isclose(moma_sol.objective_value, moma_ssq)
    assert moma_ssq < ssq
    assert np.isclose(moma_sol.objective_value,
                      moma_ref_sol.objective_value)
    assert np.isclose(moma_ssq, moma_ref_ssq)


def test_linear_moma_sanity(model, all_solvers):
    """Test optimization criterion and optimality for linear MOMA."""
    model.solver = all_solvers
    sol = model.optimize()

    with model:
        model.reactions.PFK.knock_out()
        knock_sol = model.optimize()
        sabs = (knock_sol.fluxes - sol.fluxes).abs().sum()

    with model:
        add_moma(model, linear=True)
        model.reactions.PFK.knock_out()
        moma_sol = model.optimize()
        moma_sabs = (moma_sol.fluxes - sol.fluxes).abs().sum()

    # Use normal FBA as reference solution.
    with model:
        add_moma(model, solution=sol, linear=True)
        model.reactions.PFK.knock_out()
        moma_ref_sol = model.optimize()
        moma_ref_sabs = (moma_ref_sol.fluxes - sol.fluxes).abs().sum()

    assert np.allclose(moma_sol.objective_value, moma_sabs)
    assert moma_sabs < sabs
    assert np.isclose(moma_sol.objective_value,
                      moma_ref_sol.objective_value)
    assert np.isclose(moma_sabs, moma_ref_sabs)

    with model:
        add_moma(model, linear=True)
        with pytest.raises(ValueError):
            add_moma(model)



# -*- coding: utf-8 -*-

"""Test functionalities of Geometric FBA."""

from __future__ import absolute_import

import numpy as np
import pytest
from pandas import Series

from cobra.core import Metabolite, Model, Reaction, Solution
from cobra.flux_analysis import geometric_fba


@pytest.fixture(scope="module")
def geometric_fba_model():
    """
    Generate geometric FBA model as described in [1]_

    References
    ----------
    .. [1] Smallbone, Kieran & Simeonidis, Vangelis. (2009).
           Flux balance analysis: A geometric perspective.
           Journal of theoretical biology.258. 311-5.
           10.1016/j.jtbi.2009.01.027.

    """
    test_model = Model('geometric_fba_paper_model')

    test_model.add_metabolites(Metabolite('A'))
    test_model.add_metabolites(Metabolite('B'))

    v_1 = Reaction('v1', upper_bound=1.0)
    v_1.add_metabolites({test_model.metabolites.A: 1.0})

    v_2 = Reaction('v2', lower_bound=-1000.0)
    v_2.add_metabolites({test_model.metabolites.A: -1.0,
                         test_model.metabolites.B: 1.0})

    v_3 = Reaction('v3', lower_bound=-1000.0)
    v_3.add_metabolites({test_model.metabolites.A: -1.0,
                         test_model.metabolites.B: 1.0})

    v_4 = Reaction('v4', lower_bound=-1000.0)
    v_4.add_metabolites({test_model.metabolites.A: -1.0,
                         test_model.metabolites.B: 1.0})

    v_5 = Reaction('v5')
    v_5.add_metabolites({test_model.metabolites.A: 0.0,
                         test_model.metabolites.B: -1.0})

    test_model.add_reactions([v_1, v_2, v_3, v_4, v_5])

    test_model.objective = 'v5'

    return test_model


def test_geometric_fba_benchmark(model, benchmark, all_solvers):
    """Benchmark geometric_fba."""
    model.solver = all_solvers
    benchmark(geometric_fba, model)


def test_geometric_fba(geometric_fba_model, all_solvers):
    """Test geometric_fba."""
    geometric_fba_model.solver = all_solvers
    geometric_fba_sol = geometric_fba(geometric_fba_model)
    expected = Series({'v1': 1.0, 'v2': 0.33, 'v3': 0.33, 'v4': 0.33,
                       'v5': 1.0}, index=['v1', 'v2', 'v3', 'v4', 'v5'])
    assert np.allclose(geometric_fba_sol.fluxes, expected, atol=1E-02)
